#include <SPI.h>
#include <TFT.h>
#include <Keypad.h>

#define CS 10
#define DC 9
#define RST 12
#define SUDOKU_SIZE 6
#define NUM_ELEM (SUDOKU_SIZE*SUDOKU_SIZE)

int start_boards[5][NUM_ELEM] = {
  {1, 4, 0, 0, 5, 6,
   0, 0, 0, 4, 1, 0,
   4, 2, 1, 0, 0, 0,
   6, 0, 3, 0, 4, 1,
   0, 0, 5, 0, 2, 0,
   2, 0, 0, 5, 6, 3},
   {0, 0, 1, 0, 0, 0,
   0, 5, 0, 0, 2, 4,
   0, 4, 2, 6, 0, 0,
   5, 0, 3, 0, 1, 2,
   6, 0, 5, 0, 4, 1,
   2, 1, 0, 5, 3, 0},
   {0, 5, 0, 0, 0, 6,
   2, 4, 0, 5, 0, 1,
   0, 0, 2, 0, 5, 3,
   0, 3, 0, 1, 2, 0,
   5, 6, 4, 3, 0, 0,
   0, 2, 1, 0, 6, 0},
   {1, 0, 0, 0, 2, 4,
   0, 0, 0, 1, 0, 0,
   0, 1, 0, 5, 0, 2,
   5, 2, 6, 4, 1, 3,
   0, 5, 2, 0, 3, 1,
   0, 0, 0, 0, 0, 5},
   {0, 0, 6, 0, 0, 0,
   0, 1, 0, 0, 4, 0,
   5, 2, 0, 0, 0, 0,
   6, 3, 4, 2, 5, 0,
   0, 6, 2, 5, 0, 3,
   1, 5, 3, 4, 0, 6}
};

typedef enum STATUS {MOVE, SELECT, UNFINISHED, RIGHT, WRONG};
TFT TFTscreen = TFT(CS, DC, RST);

int cell_height;
int cell_width;
byte pos_x;
byte pos_y;
byte cells_filled;
STATUS status;
byte menu_select;

int sudoku_board[NUM_ELEM];
byte blocked[NUM_ELEM];

const byte KEY_ROWS = 4;
const byte KEY_COLS = 3;
char keys[KEY_ROWS][KEY_COLS] = {
  {'1','2','3'},
  {'4','5','6'},
  {'7','8','9'},
  {'*','0','#'}
};

// byte rowPins[KEY_ROWS] = {8, 7, 6, 5};
// byte colPins[KEY_COLS] = {4, 3, 2};

byte rowPins[KEY_ROWS] = {5, 4, 3, 2};
byte colPins[KEY_COLS] = {8, 7, 6};

Keypad keypad = Keypad(makeKeymap(keys), rowPins, colPins, KEY_ROWS, KEY_COLS );

void draw_board() {
  TFTscreen.stroke(255, 255, 255);
  TFTscreen.setTextSize(1);
  int i, j;
  char text[2];
  //draw cells
  for (i = 0; i < SUDOKU_SIZE; i++) {
    for (j = 0; j < SUDOKU_SIZE; j++) {
      if (i != pos_x || j != pos_y){
        TFTscreen.rect(j * cell_width, i * cell_height, cell_width, cell_height);
        if (sudoku_board[i * SUDOKU_SIZE + j] > 0) {
          sprintf(text, "%d\0", sudoku_board[i * SUDOKU_SIZE + j]);
          TFTscreen.text(text, j * cell_width + cell_width / 3, i * cell_height + cell_height / 4);
          TFTscreen.stroke(255, 255, 255);
        }
      }
    }
  }

  //draw zones
  TFTscreen.stroke(255, 0, 255);
  TFTscreen.rect(0 * cell_width, 0 * cell_height, 3 * cell_width, 2 * cell_height);
  TFTscreen.rect(0 * cell_width, 2 * cell_height, 3 * cell_width, 2 * cell_height);
  TFTscreen.rect(0 * cell_width, 4 * cell_height, 3 * cell_width, 2 * cell_height);
  TFTscreen.rect(3 * cell_width, 0 * cell_height, 3 * cell_width, 2 * cell_height);
  TFTscreen.rect(3 * cell_width, 2 * cell_height, 3 * cell_width, 2 * cell_height);
  TFTscreen.rect(3 * cell_width, 4 * cell_height, 3 * cell_width, 2 * cell_height);
  
  //draw current position
  TFTscreen.stroke(0, 255, 255);
  TFTscreen.rect(pos_y * cell_width, pos_x * cell_height, cell_width, cell_height);
  if (sudoku_board[pos_x * SUDOKU_SIZE + pos_y] > 0) {
    sprintf(text, "%d\0", sudoku_board[pos_x * SUDOKU_SIZE + pos_y]);
    TFTscreen.text(text, pos_y * cell_width + cell_width / 3, pos_x * cell_height + cell_height / 4);
  }
  
}

void draw_menu() {
  switch(status) {
    case RIGHT:
      TFTscreen.setTextSize(2);
      TFTscreen.stroke(255, 255, 255);
      TFTscreen.text("YOU WIN!", 20, 5);
      TFTscreen.fill(255, 255, 255);
      TFTscreen.rect(10, 50, 60, 30);
      TFTscreen.stroke(0, 0, 0);
      TFTscreen.setTextSize(1);
      TFTscreen.text("NEW GAME", 17, 60);
      TFTscreen.noFill();
      break;
    default:
      TFTscreen.setTextSize(2);
      TFTscreen.stroke(255, 255, 255);
      if(status == WRONG) 
        TFTscreen.text("WRONG!", 20, 5);
      else
        TFTscreen.text("INCOMPLETE!", 20, 5);
      if (menu_select == 0) {
        TFTscreen.fill(255, 255, 255);
        TFTscreen.rect(10, 50, 60, 30);
        TFTscreen.stroke(0, 0, 0);
        TFTscreen.setTextSize(1);
        TFTscreen.text("NEW GAME", 17, 60);
        TFTscreen.noFill();
      } else {
        TFTscreen.stroke(0, 0, 0);
        TFTscreen.fill(0, 0, 0);
        TFTscreen.rect(10, 50, 60, 30);
        TFTscreen.stroke(255, 255, 255);
        TFTscreen.setTextSize(1);
        TFTscreen.text("NEW GAME", 17, 60);
        TFTscreen.noFill();
      }

      if (menu_select == 1) {
        TFTscreen.fill(255, 255, 255);
        TFTscreen.rect(87, 50, 60, 30);
        TFTscreen.stroke(0, 0, 0);
        TFTscreen.setTextSize(1);
        TFTscreen.text("CONTINUE", 93, 60);
        TFTscreen.noFill();
      } else {
        TFTscreen.stroke(0, 0, 0);
        TFTscreen.fill(0, 0, 0);
        TFTscreen.rect(87, 50, 60, 30);
        TFTscreen.stroke(255, 255, 255);
        TFTscreen.setTextSize(1);
        TFTscreen.text("CONTINUE", 93, 60);
        TFTscreen.noFill();
      }
      break;
  }
}

void set_board() {
  cells_filled = 0;
  int i, j;
  int* board = start_boards[random(0, 5)];
  for (i = 0; i < SUDOKU_SIZE; i++) {
    for (j = 0; j < SUDOKU_SIZE; j++) {
      sudoku_board[i * SUDOKU_SIZE + j] = board[i * SUDOKU_SIZE + j];
      if (sudoku_board[i * SUDOKU_SIZE + j] != 0) {
        cells_filled ++;
        blocked[i * SUDOKU_SIZE + j] = 1;
      } else {
        blocked[i * SUDOKU_SIZE + j] = 0;
      }
    }
  }
}

void validate_board() {
  if (cells_filled < NUM_ELEM){
    status = UNFINISHED;
    return;
  }
  int i, j;
  byte checked[SUDOKU_SIZE + 1];
  //check rows
  for (i = 0; i < SUDOKU_SIZE; i++) {
    memset(checked, 0, (SUDOKU_SIZE + 1) * sizeof(byte));
    for (j = 0; j < SUDOKU_SIZE; j++) {
      if (checked[sudoku_board[i * SUDOKU_SIZE + j]]) {
        status = WRONG;
        return;
      } else
        checked[sudoku_board[i * SUDOKU_SIZE + j]] = 1;
    }
  }

  //check columns
  for (i = 0; i < SUDOKU_SIZE; i++) {
    memset(checked, 0, (SUDOKU_SIZE + 1) * sizeof(byte));
    for (j = 0; j < SUDOKU_SIZE; j++) {
      if (checked[sudoku_board[j * SUDOKU_SIZE + i]]) {
        status = WRONG;
        return;
      } else
        checked[sudoku_board[j * SUDOKU_SIZE + i]] = 1;
    }
  }

  int m, n;
  //check sectors
  for (i = 0; i < 3; i++) {
    for (j = 0; j < 2; j++){
      memset(checked, 0, (SUDOKU_SIZE + 1) * sizeof(byte));
      for (m = 0; m < 2; m++) {
        for (n = 0; n < 3; n++){
          if (checked[sudoku_board[(i * 2 + m) * SUDOKU_SIZE + (j * 3 + n)]]) {
            status = WRONG;
            return;
          } else
            checked[sudoku_board[(i * 2 + m) * SUDOKU_SIZE + (j * 3 + n)]] = 1;
        }
      }
    }
  }

  status = RIGHT;
}

void setup() {
  //for debugging
  Serial.begin(9600);

  //set screen params
  TFTscreen.begin();
  TFTscreen.background(0, 0, 0);
  TFTscreen.setRotation(3);

  //set params
  cell_height = TFTscreen.height() / SUDOKU_SIZE;
  cell_width = TFTscreen.width() / SUDOKU_SIZE;
  pos_x = 0;
  pos_y = 0;
  cells_filled;
  status = MOVE;
  menu_select = 0;

  //choose a random game layout
  set_board();

  //draw the board
  draw_board();
}

void interpert_key(char key) {
  if (status == SELECT) {
	//fill cell
    if (key >= '1' && key <= '6') {
      if (sudoku_board[pos_x * SUDOKU_SIZE + pos_y] == 0)
        cells_filled++;
      sudoku_board[pos_x * SUDOKU_SIZE + pos_y] = key - '0';
      TFTscreen.fill(0, 0, 0);
      TFTscreen.stroke(255, 255, 255);
      TFTscreen.rect(pos_y * cell_width, pos_x * cell_height, cell_width, cell_height);
      TFTscreen.noFill();
      status = MOVE;
    } else if (key == '*')
      status = MOVE;
  } else if (status == MOVE) { //move on table
    switch(key) {
      case '2':
        pos_x = pos_x == 0 ? 5 : pos_x - 1;
        break;
      case '8':
        pos_x = (pos_x + 1) % SUDOKU_SIZE;
        break;
      case '4':
        pos_y = pos_y == 0 ? 5 : pos_y - 1;
        break;
      case '6':
        pos_y = (pos_y + 1) % SUDOKU_SIZE;
        break;
      case '5':
        if (!blocked[pos_x * SUDOKU_SIZE + pos_y])
          status = SELECT;
        break;
      case '#': //menu
        TFTscreen.background(0, 0, 0);
        menu_select = 0;
        validate_board();
        break;
      default: break;
    }
  } else { //menu choices selection
    switch(key) {
      case '4':
        menu_select = 0;
        break;
      case '6':
        if (status != RIGHT)
          menu_select = 1;
        break;
      case '5':
        TFTscreen.background(0, 0, 0);
        if (menu_select == 0)
          set_board();
        status = MOVE;
        break;
    }
  }
}

void loop() {
  char key = keypad.getKey();
  //key input
  if (key){
    //for debugging
    Serial.println(key);
    
    interpert_key(key);
    if (status == MOVE || status == SELECT)
      draw_board();
    else
      draw_menu();
  }
}
